/*
libKswPlugin.so,
a library to count kernel software events in VampirTrace
Copyright (C) 2010 TU Dresden, ZIH

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

 */
#include <vampirtrace/vt_plugin_cntr.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <linux/perf_event.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <pthread.h>


static int nr=0;

static int fds[1024];
static pthread_mutex_t add_counter_mutex;

int32_t init(){
  /* check if pthread mutex can be created */
  return pthread_mutex_init( &add_counter_mutex, NULL );
}


int32_t add_counter(char * event_name){
	int myNr;
  pthread_mutex_lock( &add_counter_mutex );
	myNr=nr++;
  pthread_mutex_unlock( &add_counter_mutex );
  if (myNr>=1024) return -1;
	struct perf_event_attr attr;
	memset(&attr, 0, sizeof(struct perf_event_attr));
	attr.type = PERF_TYPE_SOFTWARE;
  attr.config=PERF_COUNT_SW_MAX;
	if (strcmp (event_name, "cpu_migrations")==0)
	  attr.config=PERF_COUNT_SW_CPU_MIGRATIONS;
  if (strcmp (event_name, "task_clock")==0)
    attr.config=PERF_COUNT_SW_TASK_CLOCK;
  if (strcmp (event_name, "page_faults")==0)
    attr.config=PERF_COUNT_SW_PAGE_FAULTS;
  if (strcmp (event_name, "context_switches")==0)
    attr.config=PERF_COUNT_SW_CONTEXT_SWITCHES;
  if (attr.config==PERF_COUNT_SW_MAX){
    fprintf(stderr, "Unknown event name given:\n\t%s\n Aborting.",event_name);
    return -1;
  }
  attr.sample_period = 1;
	fds[myNr] = syscall(__NR_perf_event_open, &attr, 0, -1, -1, 0);
	if (fds[myNr]<=0){
    fprintf(stderr, "Unable to open counter %s\n Aborting.",event_name);
	  return -fds[myNr];
	}
	return myNr;
}

vt_plugin_cntr_metric_info * get_event_info(char * event_name){
	vt_plugin_cntr_metric_info * return_values;
  return_values=
			malloc(2 * sizeof(vt_plugin_cntr_metric_info) );
	return_values[0].name=strdup(event_name);
	return_values[0].unit=NULL;
	return_values[0].cntr_property=VT_PLUGIN_CNTR_LAST|VT_PLUGIN_CNTR_ACC|
     VT_PLUGIN_CNTR_UNSIGNED;
	return_values[1].name=NULL;
	return return_values;
}
uint64_t get_value(int counterIndex){
	uint64_t count=0;
	size_t res=read(fds[counterIndex], &count, sizeof(uint64_t));
	if (res!=sizeof(uint64_t))
	  return !0;
	return count;
}
void fini(){
  int i;
  for (i=0;i<nr;i++){
    close(fds[i]);
  }
}

vt_plugin_cntr_info get_info(){
	vt_plugin_cntr_info info;
  memset(&info,0,sizeof(vt_plugin_cntr_info));
	info.init              = init;
	info.add_counter       = add_counter;
  info.vt_plugin_cntr_version = VT_PLUGIN_CNTR_VERSION;
	info.run_per           = VT_PLUGIN_CNTR_PER_THREAD;
	info.synch             = VT_PLUGIN_CNTR_SYNCH;
	info.get_event_info    = get_event_info;
	info.get_current_value = get_value;
	info.finalize          = fini;
	return info;
}
